/*****************************************************************************
 * Copyright 2016-2017 Karolin Stange, Christoph Wick, and Haye Hinrichsen
 *
 * This file is part of JustIntonation.
 *
 * JustIntonation is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * JustIntonation is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along 
 * with JustIntonation. If not, see http://www.gnu.org/licenses/.
 *****************************************************************************/

//=============================================================================
//                           Audio output device
//=============================================================================

#include "audiooutputdevice.h"

#include "audiooutput.h"
#include "audiooutputdevice.h"

//-----------------------------------------------------------------------------
//                              Constructor
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Constructor, resetting the member variables
/// \param audiooutput : Pointer back to the AudioOutput
///////////////////////////////////////////////////////////////////////////////

AudioOutputDevice::AudioOutputDevice(AudioOutput *audiooutput)
    : pAudioOutput(audiooutput)
    , pQtAudioOutputStream(nullptr)
{
    setModuleName("Audio");
}


//-----------------------------------------------------------------------------
//                      Start the audio output device
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Connect and start the audio output device
/// \param parameters : AudioDeviceParameters used for the connection
/// \return True if device could be opened successfully.
///////////////////////////////////////////////////////////////////////////////

bool AudioOutputDevice::connect(const AudioParameters &parameters)
{
    if (not pAudioOutput) return false;

    // Copy the parameters. The finally realized version may be different
    AudioParameters realizedParameters(parameters);

    // If sample rate is zero set default value
    if (realizedParameters.sampleRate <= 0) realizedParameters.sampleRate = 44100;

    // If channel count is zero set default value
    if (realizedParameters.channelCount <= 0) realizedParameters.channelCount = 2;

    QAudioDeviceInfo info;
    // If device name is empty look for default device
    if (realizedParameters.deviceName.size()==0)
    {
        info = QAudioDeviceInfo::defaultOutputDevice();
        realizedParameters.deviceName = info.deviceName();
        if (realizedParameters.deviceName.size()==0)
        {
            LOGWARNING << "No default device found";
            return false;
        }
    }
    else
    {
        // Look up whether the selected device exists
        bool found = false;
        QStringList deviceList; // several devices may be compatible with the name
        QList<QAudioDeviceInfo> availableDevices(QAudioDeviceInfo::availableDevices(QAudio::AudioOutput));
        for (const QAudioDeviceInfo &deviceinfo : availableDevices)
        {
            if (deviceinfo.deviceName() == realizedParameters.deviceName)
            {
                info = deviceinfo;
                found = true;
                deviceList << deviceinfo.deviceName();
                break;
            }
        }
        if (not found)
        {
            LOGWARNING << "Could not find the audio device" << realizedParameters.deviceName
                    << "in the list" << deviceList;
            return false;
        }
    }
    LOGMESSAGE << "Trying to open" << realizedParameters.deviceName;
    LOGMESSAGE << "Tentative parameter set: samplerate =" << realizedParameters.sampleRate
            << "samplesize =" << realizedParameters.sampleSize
            << "channels =" << realizedParameters.channelCount
            << "buffersize =" << realizedParameters.bufferSize;

    // Format specification:
    QAudioFormat format;
    format.setSampleRate(realizedParameters.sampleRate);
    format.setChannelCount(realizedParameters.channelCount);
    format.setCodec("audio/pcm");
    format.setSampleSize(16);
    format.setByteOrder(QAudioFormat::LittleEndian);
    format.setSampleType(QAudioFormat::SignedInt);

    // Look up whether the format is supported
    if (not info.isFormatSupported(format))
    {
        LOGWARNING << "Audio format is not supported by backend,"
                << "falling back to preferred format";
        format = info.preferredFormat();
        if (not info.isFormatSupported(format))
        {
            LOGWARNING << "Fallback failed. Probably there is"
                    << "no audio output device available.";
            return false;
        }
    }

    LOGMESSAGE << "Format sample size and type" << format.sampleSize()
            << format.sampleType() << format.sampleRate();
    if ((format.sampleSize()!=16 and format.sampleSize()!=24)
            or format.sampleType() != QAudioFormat::SignedInt)
    {
        LOGWARNING << "Fallback sample type and/or size are not supported by the application.";
        return false;
    }

    // If the wanted sample rate and channel count
    //can not be met reset to the suggested values:
    if (format.sampleRate() != realizedParameters.sampleRate)
    {
        LOGMESSAGE << "Resetting sample rate to " << format.sampleRate();
        realizedParameters.sampleRate = format.sampleRate();
    }
    if (format.channelCount() != realizedParameters.channelCount)
    {
        LOGMESSAGE << "Resetting channel count to " << format.channelCount();
        realizedParameters.channelCount = format.channelCount();
    }

    mSampleSize = format.sampleSize();

    // ############### CREATE QT AUDIO OUTPUT STREAM ##################
    pQtAudioOutputStream = new QAudioOutput(info, format);

    // Set audio buffer size
    if (realizedParameters.bufferSize > 0)
    {
        LOGMESSAGE << "Set Qt audio buffer size to" << realizedParameters.bufferSize;
        pQtAudioOutputStream->setBufferSize(realizedParameters.bufferSize);
    }

    if (pQtAudioOutputStream->error() != QAudio::NoError)
    {
        LOGWARNING << "Error opening QAudioOutput with error " << pQtAudioOutputStream->error();
        return false;
    }

    // set volume to 100%
    pQtAudioOutputStream->setVolume(1);

    // ################ OPEN QT AUDIO OUTPUT STREAM ##################
    if (not open(QIODevice::ReadOnly))
    {
        LOGWARNING << "Audio output device could not be opened, aborting.";
        return false;
    }

    // ################ START QT AUDIO OUTPUT STREAM ##################
    pQtAudioOutputStream->start(this);
    if (pQtAudioOutputStream->error() != QAudio::NoError)
    {
        LOGWARNING << "Error opening QAudioOutput with error ",pQtAudioOutputStream->error();
        return false;
    }

    if (realizedParameters.deviceName != info.deviceName())
    {
        LOGERROR << "Consistency error";
        return false;
    }
    // Update current device name
    LOGMESSAGE << "Opened audio output device:" << realizedParameters.deviceName;
    realizedParameters.supportedSampleRates = info.supportedSampleRates();
    realizedParameters.active = true;
    pAudioOutput->setActualParameters(realizedParameters);
    return true;
}


//-----------------------------------------------------------------------------
//                         Stop the audio output device
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Stop the audio output device
///////////////////////////////////////////////////////////////////////////////

void AudioOutputDevice::disconnect()
{
    if (pQtAudioOutputStream)
    {
        pQtAudioOutputStream->stop();
        delete pQtAudioOutputStream;
        pQtAudioOutputStream = nullptr;
    }
}


//-----------------------------------------------------------------------------
//                            Set global volume
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Set the global volume
/// \param volume : Volume level between 0 and 1
///////////////////////////////////////////////////////////////////////////////

void AudioOutputDevice::setVolume(double volume)
{
    if (pQtAudioOutputStream) pQtAudioOutputStream->setVolume(volume);
}


//-----------------------------------------------------------------------------
//                              Read PCM data
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Read PCM data
/// \details This is a virtual function defined in QIODevice which is now
/// implemented here. The call is redirected to the SoundGenerator.
/// \param data : Pointer to the PCM data
/// \param maxSize : Maximal size to be generated in bytes
/// \return : Actual number of bytes that have been generated
///////////////////////////////////////////////////////////////////////////////

qint64 AudioOutputDevice::readData (char* data, qint64 maxSize)
{
    return pAudioOutput->getAudioGenerator()->generateSound(data,maxSize);
}


//-----------------------------------------------------------------------------
//                     Empty implementation for writing
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Empty implementation for writing without functionality
/// \details This function is virtual in QIODevice and has to be implemented,
/// even if it does nothing.
/// \param data
/// \param maxSize
/// \return : Zero
///////////////////////////////////////////////////////////////////////////////

qint64 AudioOutputDevice::writeData (const char * data, qint64 maxSize)
{ (void)data; (void)maxSize; return 0; }

