/*****************************************************************************
 * Copyright 2016-2017 Karolin Stange, Christoph Wick, and Haye Hinrichsen
 *
 * This file is part of JustIntonation.
 *
 * JustIntonation is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * JustIntonation is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along 
 * with JustIntonation. If not, see http://www.gnu.org/licenses/.
 *****************************************************************************/

//=============================================================================
//                         Microtonal Midi Converter
//=============================================================================

#ifndef MIDIMICROTONE_H
#define MIDIMICROTONE_H

///////////////////////////////////////////////////////////////////////////////
/// \defgroup microtonal Microtonal Midi module
/// \brief Converter for microtonal tuning of external Midi devices
/// \details The usual Midi norm was designed exclusively for the equal
/// temperament. Only the overall pitch of a voice can be controlled
/// by a so-called 'pitch-bend'. However, this affects all tones of the
/// respective channel simultaneously. Individual tuning of a single note
/// is impossible.
///
/// This restriction can be circumvented partially by the following trick.
/// Subsequent notes of the same voice can be remapped to different channels,
/// and each of these notes can then be tuned individually. Since the Midi
/// norm is restricted to at most 15 voices (plus one drum channel),
/// this means that using this trick
/// we can play no more than 15 notes together at the same time.
///
/// MidiMicrotone is a module that converts an incoming Midi stream in combination
/// with incoming dynamical tuning corrections generated by the Tuner module
/// into an outgoing reorganized microtonally tuned Midi stream.
///////////////////////////////////////////////////////////////////////////////

#include <QObject>
#include <QMutex>
#include <QMidiMessage>
#include <QTimer>

#include "midimicrotonalhelper.h"
#include "system/log.h"
#include "system/shared.h"

///////////////////////////////////////////////////////////////////////////////
/// \ingroup microtonal
/// \brief Class for converting an ordinary Midi stream into a microtonal one.
///////////////////////////////////////////////////////////////////////////////

class MidiMicrotonal : public QObject, public Log
{
    Q_OBJECT
public:
    MidiMicrotonal();

    bool init();
    bool exit();
    bool start();
    bool stop();
    void suspend();
    void resume();

    void activate (bool active);

public slots:
    ///////////////////////////////////////////////////////////////////////////
    /// \brief Input slot: Here the module receives
    /// the stream of incoming Midi events
    /// \details The input slot processes the event only if the module has
    /// been started. It is executed in the thread that instantiated the module.
    ///////////////////////////////////////////////////////////////////////////
    void receiveMidiEvent (const QMidiMessage event);

    ///////////////////////////////////////////////////////////////////////////
    /// \brief Input slot: Receive tuning corrections from the tuner (<key,cent>)
    /// \param corrections : QMap<int,double> holding the keys
    /// and the corresponding cent deviation.
    ///////////////////////////////////////////////////////////////////////////
    void receiveTuningCorrections (QMap<int,double> corrections);

    ///////////////////////////////////////////////////////////////////////////
    /// \brief Slot: Turns all notes off
    ///////////////////////////////////////////////////////////////////////////
    void allNotesOff();

    ///////////////////////////////////////////////////////////////////////////
    /// \brief Slot: Set the instrument channel of the
    /// external Midi device
    /// \details While Midi is polyphononic, the remapped Midi produced
    /// by this class can play only one instrument, which is set here.
    /// \param channel : 0...127
    ///////////////////////////////////////////////////////////////////////////
    void setMidiOutputChannel(int channel);


    ///////////////////////////////////////////////////////////////////////////
    /// \brief Slot: Re-initialize the output stream
    /// \details This slot is called when the MIDI output device changes
    ///////////////////////////////////////////////////////////////////////////
    void reInitializeConvertedMidiStream();

signals:
    ///////////////////////////////////////////////////////////////////////////
    /// \brief Output signal: This is the converted outgoing stream of Midi events
    ///////////////////////////////////////////////////////////////////////////
    void sendMidiEvent (QMidiMessage event);

    ///////////////////////////////////////////////////////////////////////////
    /// \brief onClosedMidiLoopDetected
    /// \details Signal sent when a closed Midi loop has been detected. If the
    /// application is connected to a virtual loop device (such as
    /// "Midi Through"), the emitted Midi commands immediately bounce back,
    /// leading to a short-circuit of Midi commands freezing the application.
    /// To this end the Midi output regularly sends markers which are detected
    /// on the input side. If a marker is detected, this signal is sent which
    /// should be used to disconnect the Midi output.
    ///////////////////////////////////////////////////////////////////////////
    void onClosedMidiLoopDetected();

private:

    MidiMicrotonalHelper mMidiMicrotonalHelper;
    bool mRunning;
    bool mActive;
    bool mSuspended;
    int mLastChannelModeMessage;
    QMutex mMutex;
    QTimer mTimer;

public:
    static const QMidiMessage cLoopMarker; // Marker for closed Midi loops
    bool isWorking() { return mRunning and mActive and not mSuspended; }

};

#endif // MIDIMICROTONE_H
