/*****************************************************************************
 * Copyright 2016-2017 Karolin Stange, Christoph Wick, and Haye Hinrichsen
 *
 * This file is part of JustIntonation.
 *
 * JustIntonation is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * JustIntonation is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along 
 * with JustIntonation. If not, see http://www.gnu.org/licenses/.
 *****************************************************************************/

//=============================================================================
//                          Main Tuning Algorithm
//=============================================================================

#include "tuneralgorithm.h"
#include "tuner.h"
#include "tunerdebug.h"

#include <QGenericMatrix>

//-----------------------------------------------------------------------------
//                              Constructor
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Constructor of the TunerAlgorithm
///////////////////////////////////////////////////////////////////////////////

TunerAlgorithm::TunerAlgorithm ()
    : mSelectionIndex(128,128)          // JI interval variant selection
    , mProgression(0)                   // pitch progression variable
{
    // Reset just intonation interval sizes
    mSelectionIndex.setZero();
}


//-----------------------------------------------------------------------------
//              Tune in a statically defined unequal temperament
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Tune statically in a given unequal temperament (UT)
/// \details The sampler receives tuning deviations from the tuning module
/// which are given in cents relative to the ET. Static UT's are defined by
/// a set of time-independent cent deviations which refer to a certain
/// reference key. Usually a UT works only in this reference key and its
/// complementary keys. For example, static just intonation based on the
/// reference key C sounds nice in C-Major but ugly in D-Major.
/// This function copies the given pitch deviations from the table to
/// the actual pitches of the active keys. The pitch deviations are assumed
/// ro repeat on each octave. As static temperaments do not exhibit harmonic
/// progression, the average pitch progression is zero.
/// \param keys : Array of all keys
/// \param pitches : 12 pitches defining the static temperament
/// \param referenceKey : Root key to which the temperament refers
/// \param wolfsIntervalShift : Index from which the circle of fifths starts
///////////////////////////////////////////////////////////////////////////////

void TunerAlgorithm::tuneStaticUT (KeyData &keys,
                                   const QVector<double> pitches,
                                   const int referenceKey,
                                   const int wolfsIntervalShift)
{
    for (TunerKeyData &key : keys) if (key.intensity > 0)
    {
        int distance = (key.key + 240 - referenceKey + wolfsIntervalShift)%12;
        int A4dist   = (69 + 240 - referenceKey + wolfsIntervalShift)%12;
        key.pitch = pitches[distance]-pitches[A4dist];
    }
    mProgression = 0;
}


//-----------------------------------------------------------------------------
//                  Main Tuning Algorithm: Tune dynamically
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief TunerAlgorithm::solve
/// \brief Main Tuninig Procedure: Tune dynamically
/// \details This function is at the heart of the JustIntonation project.
/// Based on the actual pattern of pressed keys, the current volume of
/// each key and an empircal memory parameter for each key, it calculates the
/// pitches for all pressed keys by solving a linear system of equations.
/// For the solution we use the open-source library 'eigen'. Most of the
/// function deals with setting up the system of linear equations A.C=B.
/// There are three contributions: (i) equations reflecting the targeted
/// interval pitch differences between the pressed keys, (ii) equations
/// expressing the tendency of an already pressed to to keep its own pitch,
/// and (iii) a single equation reflecting the constraint of maintaining
/// a predefined average pitch. The computation results are forwarded to
/// the corresponding signals in the TunerThread.
/// \param keyData : Array containing all information about the keyboard keys
/// \param intervals : Intervals according to which the tuner tunes
/// \param weights : Weights of the intervals in the least-square fit
/// \param optimizedJI : true if non-unique interval sizes shall be optimized
/// \return : Least square deviation from optimal tune
///////////////////////////////////////////////////////////////////////////////


double TunerAlgorithm::tuneDynamically (KeyData &keyData,
                              const QVector<double> intervals,
                              const QVector<double> weights,
                              bool optimizedJI)
{
    using namespace Eigen;

    // In order to limit the complexity of the tuning process, we first
    // determine the 8 most intense playing keys and the most 4 relevant
    // memorized keys.

    const int Nmax=8;

    QMultiMap<double,int> playMap,memoMap;
    for (auto &key : keyData)
    {
        if (key.intensity>0) playMap.insert(-key.intensity,key.key);
        if (key.memory>0) memoMap.insert(-key.memory,key.key);
    }

    QVector<int> playingKeys = playMap.values().mid(0,Nmax).toVector();
    const int N = playingKeys.size();
    if (N==0) return 0;
    QVector<int> memorizedKeys = memoMap.values().mid(0,Nmax-N).toVector();
    const int M = N + memorizedKeys.size();
    // Sorting the keys is not necessary, but convenient for debugging:
    qSort(playingKeys); qSort(memorizedKeys);
    QVector<int> allKeys(playingKeys); allKeys << memorizedKeys;

    //qDebug() << "N =" << N << playingKeys << "   M-N =" << M-N << memorizedKeys;

    // Copy pitch, intensity and memory level
    VectorXd pitch(M), significance(M); bool newKeys = false;
    for (int i=0; i<M; ++i)
    {
        TunerKeyData& entry = keyData[allKeys[i]];
        if (entry.newlyPressed) newKeys = true;
        pitch(i) = entry.pitch;
        significance(i) = entry.intensity + entry.memory;
    }

    // Setting up various auxiliary matrices
    MatrixXi variants(M,M), semitones(M,M), direction(M,M);
    MatrixXd interval(M,M), weight(M,M);
    for (int i=0; i<M; ++i) for (int j=0; j<M; ++j)
    {
        semitones(i,j) = abs(allKeys[j] - allKeys[i]);
        direction(i,j) = (allKeys[j] >= allKeys[i] ? 1:-1);
        variants(i,j)  = cJustIntonation[semitones(i,j)%12].size();        
        int variant = mSelectionIndex(allKeys[i],allKeys[j]);
        interval(i,j) = direction(i,j) *
                (optimizedJI ? cJustIntonation[semitones(i,j)%12][variant]
                             : intervals[semitones(i,j)%12]);
        if (j>=N) interval(i,j) -= pitch(j);
        if (i>=N) interval(i,j) += pitch(i);
        weight(i,j)    = (i==j or (i>=N and j>=N) ? 0 : 1) *
                         (i>=N or j>=N ? memoryGlobalWeight : 1) *
                         weights[semitones(i,j)%12] *
                         std::pow(octaveWeightFactor, (semitones(i,j)-1)/12) *
                         sqrt(significance(i) * significance(j));
    }

    // Construct the matrix A and its inverse AI
    MatrixXd A = -weight.block(0,0,N,N) + MatrixXd::Identity(N,N)*epsilon
               + MatrixXd(VectorXd(weight.array().rowwise().sum()).head(N).asDiagonal());
    MatrixXd AI = A.inverse();

    // ######################## BINARY SEARCH ############################

    if (optimizedJI and newKeys)
    {
//        qDebug() << "--------------------------------------";
//         for (int i=0; i<M; ++i)
//         {
//             QString line;
//             for (int j=0; j<M; ++j) line += QString::number(weight(i,j)) + "   ";
//             qDebug() << line;
//         }
//         qDebug() << "--------------------------------------";
        VectorXd U(N); U.setZero();
        MatrixXi m(M,M), mmin(M,M); m.setZero(); mmin.setZero();
        for (int i=0; i<M; i++) for (int j=0; j<M; j++)
        {
            if (keyData[allKeys[i]].newlyPressed or keyData[allKeys[j]].newlyPressed) mmin(j,i)=mmin(i,j)=0;
            else m(i,j) = mSelectionIndex(allKeys[i],allKeys[j]);
        }
        bool searching = true;
        double Pmin = 1e100;
        while (searching)
        {
            MatrixXd interval(M,M);
            for (int i=0; i<M; ++i) for (int j=0; j<M; ++j)
            {
                interval(i,j) = direction(i,j) * cJustIntonation[semitones(i,j)%12][m(i,j)];
                if (j>=N) interval(i,j) -= pitch(j);
                if (i>=N) interval(i,j) += pitch(i);
            }
            VectorXd B = (interval * weight).diagonal().head(N) - epsilon*pitch.head(N);
            VectorXd V = -AI * B;
            double C = (interval.array() * interval.array() * weight.array()).sum() / 4;
            double P = V.dot(A*V)/2 + B.dot(V) + C;
            if (P<Pmin-1E-7) { Pmin=P; mmin=m; U=V; }

            //increment
            searching = false;
            for (int i=0; i<M; ++i) for (int j=0; j<M; ++j) if (i<j) if (keyData[allKeys[i]].newlyPressed or keyData[allKeys[j]].newlyPressed) if (variants(i,j)>1 and not searching)
            {
                m(j,i) = m(i,j) = (m(i,j)+1) % variants(i,j);
                if (m(i,j)>0) searching = true;
            }
        }
//        qDebug() << "var" << f(variants);
//        qDebug() << "mmin" << f(mmin);
//        qDebug() << "U" << f(U);
        for (int i=0; i<M; i++) for (int j=0; j<M; j++)
            mSelectionIndex(allKeys[i],allKeys[j]) = mmin(i,j);
        for (int i=0; i<N; ++i) keyData[playingKeys[i]].newlyPressed = false;
    }

    // Tune the keys micrononally:
    VectorXd B = (interval * weight).diagonal().head(N) - epsilon*pitch.head(N);
    VectorXd U = - AI * B;
    for (int i=0; i<N; i++) keyData[playingKeys[i]].pitch = U(i);

    // Compute potential as a measure of how much the chord is tempered
    double C = (interval.array() * interval.array() * weight.array()).sum() / 4;
    const double V = U.dot(A*U)/2 + B.dot(U) + C;
    return V;
}
