/*****************************************************************************
 * Copyright 2016-2017 Karolin Stange, Christoph Wick, and Haye Hinrichsen
 *
 * This file is part of JustIntonation.
 *
 * JustIntonation is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * JustIntonation is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along 
 * with JustIntonation. If not, see http://www.gnu.org/licenses/.
 *****************************************************************************/

//=============================================================================
//                  Class for playing a startup sound
//=============================================================================

#include "startupsound.h"

#include <QMediaPlayer>
#include <QSettings>
#include <QStandardPaths>
#include <QFile>
#include <QDir>
#include <QResource>

#define STATUS  if (mVerbosity >= 4) qDebug() << "StartupSound:"
#define MESSAGE if (mVerbosity >= 3) qDebug() << "StartupSound:"
#define WARNING if (mVerbosity >= 2) qWarning()<<"StartupSound: WARNING:"
#define ERROR   if (mVerbosity >= 1) qWarning() << "StartupSound: ERROR:"

//-----------------------------------------------------------------------------
//                                Constructor
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Constructor
/// \param resourcefile : Location of the resource audio file
/// \see config.h
///////////////////////////////////////////////////////////////////////////////

StartupSound::StartupSound(const QString &resourcefile)
    : mResourceFile(resourcefile)
    , mVerbosity(2)
{}


//-----------------------------------------------------------------------------
//                     Ask whether startup sound is enabled
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief StartupSound::isEnabled
/// \return True if enabled, false otherwise
///////////////////////////////////////////////////////////////////////////////

bool StartupSound::isEnabled()
{
    QSettings settings;
    return settings.value("app/startupsound",true).toBool();
}


//-----------------------------------------------------------------------------
//                 Public slot: Enable / disable startup sound
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Public slot: Enable / disable startup sound
/// \details This function enables or disables the startup sounds and stores
/// this information in the QSettings.
/// \param enabled : True if startup sound is enabled, false otherwise
///////////////////////////////////////////////////////////////////////////////

void StartupSound::enable (bool enabled)
{
    QSettings settings;
    STATUS << "Setting enabled =" << (enabled ? "true" : "false");
    settings.setValue("app/startupsound",enabled);
}


//-----------------------------------------------------------------------------
//                            Set verbosity level
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Set verbosity level of qDebug() messages.
/// \details The levels are:
/// - 0: no messages
/// - 1: errors only
/// - 2: errors and warnings
/// - 3: in addition non-repeating messages (e.g. at startup)
/// - 4: all messages
///////////////////////////////////////////////////////////////////////////////

void StartupSound::setVerbosity(int level)
{
    qDebug() << "StartupSound: Setting verbosity of messages to level" << level;
    mVerbosity = level;
}

//-----------------------------------------------------------------------------
//                              Play startup sound
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////
/// \brief Play startup sound.
/// \details This function plays the startup sound from a resource that was
/// specified when calling the constructor. A QMediaPlayer is created
/// dynamically in the heap and it is not deleted. Therefore, do not use this
/// function repeatedly for recurrent sounds.
///////////////////////////////////////////////////////////////////////////////

void StartupSound::play()
{
    if (not isEnabled())
    {
        MESSAGE << "Startup sound is disabled in the settings";
        return;
    }
    MESSAGE << "Play startup sound" << mResourceFile;

    // Identify the extension of the file (mp3,ogg,...)
    if (mResourceFile.size()==0) return;
    int lastPoint = mResourceFile.lastIndexOf(".");
    const QString extension =
            mResourceFile.right(mResourceFile.size()-lastPoint-1);
    if (extension.size() != 3)
    {
        WARNING << "Could not identify extension of resource file"
                << mResourceFile << ":" << extension;
        return;
    }

    // MAC-OS cannot play the startupsound directly from the resource file
    // For this reason the mp3/ogg file is first copied to local disk
    // First make sure that the local storage is writable
    const QString localpath =
          QStandardPaths::writableLocation(QStandardPaths::AppDataLocation);
    QDir directory;
    if (not directory.mkpath(localpath))
    {
        WARNING << "Cannot create local path" << localpath;
        return;
    }

    // Then copy the file
    const QString filename = "startupsound." + extension;
    QFile localfile (localpath + "/" + filename);
    if (localfile.exists())
    {
        if (localfile.size() != QResource(mResourceFile).size())
        {
            // If file size changed delete and reload
            localfile.remove();
            QFile::copy(mResourceFile, localfile.fileName());
        }
    }
    else QFile::copy(mResourceFile, localfile.fileName());
    if (not localfile.exists()) WARNING<< "Startupsound could not be copied.";
    STATUS << mResourceFile << "was copied to" << localfile.fileName();

    QMediaPlayer *player = new QMediaPlayer(this);
    player->setMedia(QUrl::fromLocalFile(localfile.fileName()));
    player->setVolume(50);
    player->play();
    if (player->error() != QMediaPlayer::NoError)
        WARNING << "Startupsound:" << player->errorString().toStdString().c_str();
    STATUS << "Leaving Startupsound::play()";
}

